{-# OPTIONS -fglasgow-exts #-}

module Traversal (Coverage(..), VD(..), HD(..),
                  Trans(..), Accum(..), AccTrans(..), 
                  Ctx(..), NoCtx(..),
                  GenU, GenA, GenT, GenAT, GenCA, GenCT, GenCAT,
                  mk, always, cond, whenever, malways, mcond, mwhenever,
                  traverse, xmost, innermost, outermost)
    where

import Data.Generics
import Maybe
import Core

--------------------------------------------------------------------------------
-- Data types for describing different aspects of a traversal
--------------------------------------------------------------------------------

data Trans = Trans
data Accum = Accum
data AccTrans = AccTrans

data Ctx = Ctx 
data NoCtx = NoCtx

data Coverage = Full | Spine | Once | Stop 

data VD = FromTop | FromBottom 

data HD = FromLeft | FromRight 

--------------------------------------------------------------------------------
-- Recursive traversals
--------------------------------------------------------------------------------

-- Five aspects of a one-pass traversal:
--
-- 1. coverage: full, spine, once, or stop
-- 2. vertical direction: topdown or bottomup
-- 3. horizontal direction: left-to-right or right-to-left
-- 4. context: contextual or non-contextual
-- 5. visit: a transformer, an accumulator, or an accumulative transformer

class Traversal u v c a t x | u v c a t -> x where
  traverse :: u -> v -> Coverage -> VD -> HD -> x

-- "mother g f" yields:

--               top-down        bottom-up
-- +-------++-------+------+  +-------+------+
-- | f\g   ||   ;   |   +  |  |   ;'  |   +' |
-- +=======++=======+======+  +=======+======+
-- | all_l || full  | stop |  | full  | stop |
-- +-------++-------+------+  +-------+------+ left-to-right
-- | one_l || spine | once |  | spine | once |
-- +-------++-------+------+  +-------+------+
-- +-------++-------+------+  +-------+------+
-- | all_r || full  | stop |  | full  | stop |
-- +-------++-------+------+  +-------+------+ right-to-left
-- | one_r || spine | once |  | spine | once |
-- +-------++-------+------+  +-------+------+

travt :: Coverage -> VD -> HD -> GCU c -> GCAT c a -> GCAT c a
travt cov v h = mother (g cov v) (f cov h)
    where g :: Coverage -> VD -> GAT a -> GAT a -> GAT a
          g Full  FromTop = compose
          g Spine FromTop = compose
          g Once  FromTop = choice
          g Stop  FromTop = choice
          g Full  FromBottom = compose'
          g Spine FromBottom = compose'
          g Once  FromBottom = choice'
          g Stop  FromBottom = choice'
          f :: Coverage -> HD -> GAT a -> GAT a
          f Full  FromLeft  = all_l
          f Stop  FromLeft  = all_l 
          f Spine FromLeft  = one_l
          f Once  FromLeft  = one_l
          f Full  FromRight = all_r
          f Stop  FromRight = all_r 
          f Spine FromRight = one_r
          f Once  FromRight = one_r

-- default behavior of a visit. it is used as a "catch-all", which is applied
-- when the specific visit(s) is not applicable.

catchv :: Coverage -> GAT a
catchv Full  = vsucc
catchv Spine = vsucc
catchv Once  = vfail
catchv Stop  = vfail

instance Data t => Traversal AccTrans NoCtx c a t
    ([GenAT c a] -> a -> t -> Maybe (a,t)) where
    traverse _ _ cov vd hd vs = 
        (travt cov vd hd) defcu (foldCAT (catchv cov) (map f vs)) defc
        where f (GenAT v) = GenCAT (\c -> v)

instance Data t => Traversal Trans NoCtx c a t
    ([GenT c a] -> t -> Maybe t) where
    traverse _ _ cov vd hd vs = \t ->
        msnd $ (travt cov vd hd) defcu (foldCAT (catchv cov) (map f vs)) defc defa t
        where f (GenT v) = GenCAT (\c a t -> defa `mpairl` v t)

instance Data t => Traversal Accum NoCtx c a t
    ([GenA c a] -> a -> t -> Maybe a) where
    traverse _ _ cov vd hd vs = \a t ->
        mfst $ (travt cov vd hd) defcu (foldCAT (catchv cov) (map f vs)) defc a t
        where f (GenA v) = GenCAT (\c a t -> v a t `mpairr` t)

instance Data t => Traversal AccTrans Ctx c a t
    ([GenU c a] -> [GenCAT c a] -> 
     (c -> a -> t -> Maybe (a,t))) where
    traverse _ _ cov vd hd us vs = 
        (travt cov vd hd) (foldC us) (foldCAT (catchv cov) vs)

instance Data t => Traversal Trans Ctx c a t
    ([GenU c a] -> [GenCT c a] -> c -> t -> Maybe t) where
    traverse _ _ cov vd hd us vs = \c t -> 
        msnd $ (travt cov vd hd) (foldC us) (foldCAT (catchv cov) (map f vs)) c defa t
        where f (GenCT v) = GenCAT (\c a t -> defa `mpairl` v c t)

instance Data t => Traversal Accum Ctx c a t
    ([GenU c a] -> [GenCA c a] -> c -> a -> t -> Maybe a) where
    traverse _ _ cov vd hd us vs = \c a t ->
        mfst $ (travt cov vd hd) (foldC us) (foldCAT (catchv cov) (map f vs)) c a t
        where f (GenCA v) = GenCAT (\c a t -> v c a t `mpairr` t)

--------------------------------------------------------------------------------
-- Fixpoint traversals: innermost/outermost
--------------------------------------------------------------------------------

-- Four aspects of an innermost/outermost traversal:
--
-- 1. vertical direction: topdown (outermost) or bottomup (innermost)
-- 2. horizontal direction: left-to-right or right-to-left
-- 3. context: contextual or non-contextual
-- 4. visit: a transformer, an accumulator, or an accumulative transformer

class Xmost u v c a t x | u v c a t -> x where
     xmost :: u -> v -> VD -> HD -> x

outermost u v = xmost u v FromTop
innermost u v = xmost u v FromBottom

-- +-------++-----------+-----------+
-- |   g   ||     +     |     +'    |
-- +=======++===========+===========+
-- | xm g  || innermost | outermost |
-- +-------++-----------+-----------+

mostt :: VD -> HD -> GCU c -> GCAT c a -> GCAT c a
mostt vd hd = xm (g vd) (f hd)
    where g :: VD -> GAT a -> GAT a -> GAT a
          g FromTop = choice
          g FromBottom = choice'
          f :: HD -> GAT a -> GAT a
          f FromLeft = one_l
          f FromRight = one_r

instance Data t => Xmost AccTrans NoCtx c a t
    ([GenAT c a] -> a -> t -> (a,t)) where
    xmost _ _ vd hd vs a t = fromJust $ 
        mostt vd hd defcu (foldCAT vfail (map f vs)) defc a t
        where f (GenAT v) = GenCAT (\c -> v)

instance Data t => Xmost Trans NoCtx c a t
    ([GenT c a] -> t -> t) where
    xmost _ _  vd hd vs t = snd $ fromJust $
        mostt vd hd defcu (foldCAT vfail (map f vs)) defc defa t
        where f (GenT v) = GenCAT (\c a t -> defa `mpairl` v t)

instance Data t => Xmost Accum NoCtx c a t
    ([GenA c a] -> a -> t -> a) where
    xmost _ _  vd hd vs a t = fst $ fromJust $
        mostt vd hd defcu (foldCAT vfail (map f vs)) defc a t
        where f (GenA v) = GenCAT (\c a t -> v a t `mpairr` t)

instance Data t => Xmost AccTrans Ctx c a t
    ([GenU c a] -> [GenCAT c a] -> c -> a -> t -> (a,t)) where
    xmost _ _ vd hd us vs c a t = fromJust $
        mostt vd hd (foldC us) (foldCAT vfail vs) c a t

instance Data t => Xmost Trans Ctx c a t
    ([GenU c a] -> [GenCT c a] -> c -> t -> t) where
    xmost _ _ vd hd us vs c t = snd $ fromJust $
        mostt vd hd (foldC us) (foldCAT vfail (map f vs)) c defa t
        where f (GenCT v) = GenCAT (\c a t -> defa `mpairl` v c t)

instance Data t => Xmost Accum Ctx c a t
    ([GenU c a] -> [GenCA c a] -> c -> a -> t -> a) where
    xmost _ _ vd hd us vs c a t = fst $ fromJust $
        mostt vd hd (foldC us) (foldCAT vfail (map f vs)) c a t
        where f (GenCA v) = GenCAT (\c a t -> v c a t `mpairr` t)


--------------------------------------------------------------------------------
-- Default values for accumulation and context, when they are absent
--------------------------------------------------------------------------------
-- default context updater
defcu :: GCU c
defcu = undefined

-- default context
defc :: c
defc = undefined

-- default accumulation
defa :: a
defa = undefined

--------------------------------------------------------------------------------
-- Auxiliary data type for encapsulating specific context updaters/visits so
-- that they can be put into lists
--------------------------------------------------------------------------------
data GenU c a   = forall t. Typeable t => GenU (c -> t -> c)

data GenCAT c a = forall t. Typeable t => GenCAT (c -> a -> t -> Maybe (a,t))
data GenCT c a  = forall t. Typeable t => GenCT (c -> t -> Maybe t)
data GenCA c a  = forall t. Typeable t => GenCA (c -> a -> t -> Maybe a)

data GenAT c a  = forall t. Typeable t => GenAT (a -> t -> Maybe (a,t))
data GenT c a   = forall t. Typeable t => GenT (t -> Maybe t)
data GenA c a   = forall t. Typeable t => GenA (a -> t -> Maybe a)

--------------------------------------------------------------------------------
-- Overload "mke" function for the encapsulation
--------------------------------------------------------------------------------
class MK c a t s x | x c a t -> s where
    mke :: s -> x c a

instance Typeable t => MK c a t (c -> t -> c) GenU where
    mke = GenU

instance Typeable t => MK c a t (a -> t -> Maybe (a,t)) GenAT where
    mke = GenAT
instance Typeable t => MK c a t (t -> Maybe t) GenT where
    mke = GenT
instance Typeable t => MK c a t (a -> t -> Maybe a) GenA where
    mke = GenA

instance Typeable t => MK c a t (c -> a -> t -> Maybe (a,t)) GenCAT where
    mke = GenCAT
instance Typeable t => MK c a t (c -> t -> Maybe t) GenCT where
    mke = GenCT
instance Typeable t => MK c a t (c -> a -> t -> Maybe a) GenCA where
    mke = GenCA

mk x = [mke x]

--------------------------------------------------------------------------------
-- The tiny embedded language for composing visits
--------------------------------------------------------------------------------

class MK c a t s x => Comb c a t u s p x | c a t x -> u s p where
    malways :: s -> [x c a]
    mcond :: p -> s -> s -> [x c a]
    mwhenever :: s -> p -> [x c a]

    always :: u -> [x c a]
    cond :: p -> u -> u -> [x c a]
    whenever :: u -> p -> [x c a]

instance Typeable t => Comb c a t 
                            (a -> t -> (a,t)) 
                            (a -> t -> Maybe (a,t)) 
                            (a -> t -> Bool) 
                            GenAT
    where
    malways = mk
    mcond p f g = malways (\a t -> if p a t then f a t else g a t)
    mwhenever f p = malways (\a t -> if p a t then f a t else Nothing)
    always f = malways (\a t -> Just $ f a t)
    cond p f g = malways (\a t -> Just $ if p a t then f a t else g a t)
    whenever f p = malways (\a t -> if p a t then Just (f a t) else Nothing)

instance Typeable t => Comb c a t 
                            (t -> t) 
                            (t -> Maybe t)
                            (t -> Bool) 
                            GenT
    where
    malways = mk
    mcond p f g = malways (\t -> if p t then f t else g t)
    mwhenever f p = malways (\t -> if p t then f t else Nothing)
    always f = malways (\t -> Just $ f t)
    cond p f g = malways (\t -> Just $ if p t then f t else g t)
    whenever f p = malways (\t -> if p t then Just (f t) else Nothing)

instance Typeable t => Comb c a t 
                            (a -> t -> a) 
                            (a -> t -> Maybe a) 
                            (a -> t -> Bool) 
                            GenA
    where
    malways = mk
    mcond p f g = malways (\a t -> if p a t then f a t else g a t)
    mwhenever f p = malways (\a t -> if p a t then f a t else Nothing)
    always f = malways (\a t -> Just $ f a t)
    cond p f g = malways (\a t -> Just $ if p a t then f a t else g a t)
    whenever f p = malways (\a t -> if p a t then Just (f a t) else Nothing)

instance Typeable t => Comb c a t 
                            (c -> a -> t -> (a,t)) 
                            (c -> a -> t -> Maybe (a,t)) 
                            (c -> a -> t -> Bool) 
                            GenCAT
    where
    malways = mk
    mcond p f g = malways (\c a t -> if p c a t then f c a t else g c a t)
    mwhenever f p = malways (\c a t -> if p c a t then f c a t else Nothing)
    always f = malways (\c a t -> Just $ f c a t)
    cond p f g = malways (\c a t -> Just $ if p c a t then f c a t else g c a t)
    whenever f p = malways (\c a t -> if p c a t then Just (f c a t) else Nothing)

instance Typeable t => Comb c a t 
                            (c -> t -> t) 
                            (c -> t -> Maybe t) 
                            (c -> t -> Bool) 
                            GenCT
    where
    malways = mk
    mcond p f g = malways (\c t -> if p c t then f c t else g c t)
    mwhenever f p = malways (\c t -> if p c t then f c t else Nothing)
    always f = malways (\c t -> Just $ f c t)
    cond p f g = malways (\c t -> Just $ if p c t then f c t else g c t)
    whenever f p = malways (\c t -> if p c t then Just (f c t) else Nothing)

instance Typeable t => Comb c a t 
                            (c -> a -> t -> a) 
                            (c -> a -> t -> Maybe a) 
                            (c -> a -> t -> Bool) 
                            GenCA
    where
    malways = mk
    mcond p f g = malways (\c a t -> if p c a t then f c a t else g c a t)
    mwhenever f p = malways (\c a t -> if p c a t then f c a t else Nothing)
    always f = malways (\c a t -> Just $ f c a t)
    cond p f g = malways (\c a t -> Just $ if p c a t then f c a t else g c a t)
    whenever f p = malways (\c a t -> if p c a t then Just (f c a t) else Nothing)

--------------------------------------------------------------------------------
-- Helper functions for constructing generic context updaters and visits
--------------------------------------------------------------------------------

newtype Xext c a t = Xext {unXext :: c -> a -> t -> Maybe (a,t)}

extCAT :: (Typeable t1, Typeable t2) => 
          (c -> a -> t1 -> Maybe (a,t1)) ->
	  (c -> a -> t2 -> Maybe (a,t2)) ->
	  c -> a -> t1 -> Maybe (a,t1)
extCAT def ext = unXext ((Xext def) `ext0` (Xext ext))

extC :: (Typeable t1, Typeable t2) =>
	(c -> t1 -> c) ->
	(c -> t2 -> c) ->
	(c -> t1 -> c)
extC f g c t = maybe (f c t) (g c) (cast t)

foldC :: [GenU c a] -> GCU c
foldC xs = foldl ext const xs
    where ext f (GenU g) = f `extC` g

foldCAT :: GAT a -> [GenCAT c a] -> GCAT c a
foldCAT d xs = foldl ext (\_ -> d) xs
    where ext f (GenCAT g) = f `extCAT` g

--------------------------------------------------------------------------------
-- Helper functions for extracting and injecting tuples in monads
--------------------------------------------------------------------------------
mmap :: Monad m => (a -> b) -> m a -> m b
mmap f x = x >>= return . f

msnd :: Monad m => m (a,b) -> m b
msnd = mmap snd

mfst :: Monad m => m (a,b) -> m a
mfst = mmap fst

mpairl :: Monad m => a -> m b -> m (a,b)
mpairl x = mmap ((,) x)

mpairr :: Monad m => m a -> b -> m (a, b)
mpairr z x = mmap (flip (,) x) z

